/*-
 * Copyright (c) 1999 James Howard and Dag-Erling Codan Smrgrav
 * Copyright (C) 2008 Gabor Kovesdan <gabor@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>

#ifndef lint
#if 0
__RCSID("$OpenBSD: grep.c,v 1.39 2007/09/02 15:19:32 deraadt Exp $");
#else
__FBSDID("$FreeBSD$");
#endif
#endif /* not lint */

#include <sys/limits.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <getopt.h>
#include <libgen.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "grep.h"

#ifndef WITHOUT_NLS
#include <nl_types.h>
nl_catd	 catalog;
#endif

/*
 * Default messags to use when NLS is disabled or no catalogue
 * is found.
 */
char	*errstr[] = {
	"",
/* 1*/	"(standard input)",
/* 2*/	"(fd %d)",
/* 3*/	"invalid file type",
/* 4*/	"usage: %s [-abcDEFGHhIiJLlmnOoPqRSsUVvwxZ] [-A num] [-B num] [-C[num]]\n",
/* 5*/	"\t[-e pattern] [-f file] [--binary-files=value] [--color=when]\n",
/* 6*/	"\t[--context[=num]] [--directories=action] [--label] [--line-buffered]\n",
/* 7*/	"\t[--null] [pattern] [file ...]\n",
/* 8*/	"parentheses not balanced",
/* 9*/	"context out of range",
/*10*/	"%s (BSD grep) %s\n",
/*11*/	"unknown --binary-files option",
/*12*/	"Binary file %s matches\n",
/*13*/	"value out of range",
/*14*/	"unknown -d or --directory option",
/*15*/	"unknown --color option",
/*16*/	"cannot read bzip2 compressed file"
};

/* Flags passed to regcomp() and regexec() */
int		 cflags;
int		 eflags = REG_STARTEND;

/* Shortcut for matching all cases like empty regex */
int		 matchall;

/* Searching patterns */
int		 patterns, pattern_sz;
char		**pattern;
regex_t		*r_pattern;
#ifdef WITH_PCRE
pcre		**perl_pattern;
#endif

/* Filename exclusion/inclusion patterns */
int		 epatterns, epattern_sz;
char		**epattern;

/* For regex errors  */
char	 re_error[RE_ERROR_BUF + 1];

/* Command-line flags */
unsigned long long Aflag;	/* -A x: print x lines trailing each match */
unsigned long long Bflag;	/* -B x: print x lines leading each match */
int	 Hflag;		/* -H: always print file name */
int	 Lflag;		/* -L: only show names of files with no matches */
int	 bflag;		/* -b: show block numbers for each match */
int	 cflag;		/* -c: only show a count of matching lines */
int	 hflag;		/* -h: don't print filename headers */
int	 iflag;		/* -i: ignore case */
int	 lflag;		/* -l: only show names of files with matches */
int	 mflag;		/* -m x: stop reading the files after x matches */
unsigned long long mcount;	/* count for -m */
int	 nflag;		/* -n: show line numbers in front of matching lines */
int	 oflag;		/* -o: print only matching part */
int	 qflag;		/* -q: quiet mode (don't output anything) */
int	 sflag;		/* -s: silent mode (ignore errors) */
int	 vflag;		/* -v: only show non-matching lines */
int	 wflag;		/* -w: pattern must start and end on word boundaries */
int	 xflag;		/* -x: pattern must match entire line */
int	 lbflag;	/* --line-buffered */
int	 nullflag;	/* --null */
int	 exclflag;	/* --exclude */
int	 inclflag;	/* --include */
char	*label;		/* --label */
char	*color;		/* --color */
int	 grepbehave = GREP_BASIC;	/* -EFGP: type of the regex */
int	 binbehave = BINFILE_BIN;	/* -aIU: handling of binary files */
int	 filebehave = FILE_STDIO;	/* -JZ: normal, gzip or bzip2 file */
int	 devbehave = DEV_GREP;		/* -D: handling of devices */
int	 dirbehave = DIR_GREP;		/* -dRr: handling of directories */
int	 linkbehave = LINK_GREP;	/* -OpS: handling of symlinks */

enum {
	BIN_OPT = CHAR_MAX + 1,
	COLOR_OPT,
	HELP_OPT,
	MMAP_OPT,
	LINEBUF_OPT,
	LABEL_OPT,
	NULL_OPT,
	R_EXCLUDE_OPT,
	R_INCLUDE_OPT
};

/* Housekeeping */
int	 first = 1;	/* flag whether we are processing the first match */
int	 prev;		/* flag whether or not the previous line matched */
int	 tail;		/* lines left to print */
int	 notfound;	/* file not found */

extern char	*__progname;

/*
 * Prints usage information and returns 2.
 */
static void
usage(void)
{
	fprintf(stderr, getstr(4), __progname);
	fprintf(stderr, getstr(5));
	fprintf(stderr, getstr(5));
	fprintf(stderr, getstr(6));
	fprintf(stderr, getstr(7));
	exit(2);
}

static char	*optstr = "0123456789A:B:C:D:EFGHIJLOPSRUVZabcd:e:f:hilm:nopqrsuvwxy";

struct option long_options[] =
{
	{"binary-files",	required_argument,	NULL, BIN_OPT},
	{"help",		no_argument,		NULL, HELP_OPT},
	{"mmap",		no_argument,		NULL, MMAP_OPT},
	{"line-buffered",	no_argument,		NULL, LINEBUF_OPT},
	{"label",		required_argument,	NULL, LABEL_OPT},
	{"null",		no_argument,		NULL, NULL_OPT},
	{"color",		optional_argument,	NULL, COLOR_OPT},
	{"colour",		optional_argument,	NULL, COLOR_OPT},
	{"exclude",		required_argument,	NULL, R_EXCLUDE_OPT},
	{"include",		required_argument,	NULL, R_INCLUDE_OPT},
	{"after-context",	required_argument,	NULL, 'A'},
	{"text",		no_argument,		NULL, 'a'},
	{"before-context",	required_argument,	NULL, 'B'},
	{"byte-offset",		no_argument,		NULL, 'b'},
	{"context",		optional_argument,	NULL, 'C'},
	{"count",		no_argument,		NULL, 'c'},
	{"devices",		required_argument,	NULL, 'D'},
        {"directories",		required_argument,	NULL, 'd'},
	{"extended-regexp",	no_argument,		NULL, 'E'},
	{"regexp",		required_argument,	NULL, 'e'},
	{"fixed-strings",	no_argument,		NULL, 'F'},
	{"file",		required_argument,	NULL, 'f'},
	{"basic-regexp",	no_argument,		NULL, 'G'},
	{"no-filename",		no_argument,		NULL, 'h'},
	{"with-filename",	no_argument,		NULL, 'H'},
	{"ignore-case",		no_argument,		NULL, 'i'},
	{"bz2decompress",	no_argument,		NULL, 'J'},
	{"files-with-matches",	no_argument,		NULL, 'l'},
	{"files-without-match", no_argument,            NULL, 'L'},
	{"max-count",		required_argument,	NULL, 'm'},
	{"line-number",		no_argument,		NULL, 'n'},
	{"only-matching",	no_argument,		NULL, 'o'},
	{"perl-regexp",		no_argument,		NULL, 'P'},
	{"quiet",		no_argument,		NULL, 'q'},
	{"silent",		no_argument,		NULL, 'q'},
	{"recursive",		no_argument,		NULL, 'r'},
	{"no-messages",		no_argument,		NULL, 's'},
	{"binary",		no_argument,		NULL, 'U'},
	{"unix-byte-offsets",	no_argument,		NULL, 'u'},
	{"invert-match",	no_argument,		NULL, 'v'},
	{"version",		no_argument,		NULL, 'V'},
	{"word-regexp",		no_argument,		NULL, 'w'},
	{"line-regexp",		no_argument,		NULL, 'x'},
	{"decompress",          no_argument,            NULL, 'Z'},
	{NULL,			no_argument,		NULL, 0}
};

/*
 * Adds a searching pattern to the internal array.
 */
static void
add_pattern(char *pat, size_t len)
{

	/* Check if we can do a shortcut */
	if (len == 0 || matchall) {
		matchall = 1;
		return;
	}
	/* Increase size if necessary */
	if (patterns == pattern_sz) {
		pattern_sz *= 2;
		pattern = grep_realloc(pattern, ++pattern_sz * sizeof(*pattern));
	}
	if (len > 0 && pat[len - 1] == '\n')
		--len;
	/* pat may not be NUL-terminated */
	if (wflag && !(grepbehave == GREP_FIXED)) {
		int bol = 0, eol = 0, extra;
		if (pat[0] == '^')
			bol = 1;
		if (len > 0 && pat[len - 1] == '$')
			eol = 1;
		extra = (grepbehave == GREP_EXTENDED) ? 2 : 4;
		pattern[patterns] = grep_malloc(len + 15 + extra);
		snprintf(pattern[patterns], len + 15 + extra,
		   "%s[[:<:]]%s%.*s%s[[:>:]]%s",
		    bol ? "^" : "",
		    (grepbehave == GREP_EXTENDED) ? "(" : "\\(",
		    (int)len - bol - eol, pat + bol,
		    (grepbehave == GREP_EXTENDED) ? ")" : "\\)",
		    eol ? "$" : "");
		len += 14 + extra;
	} else {
		pattern[patterns] = grep_malloc(len + 1);
		memcpy(pattern[patterns], pat, len);
		pattern[patterns][len] = '\0';
	}
	++patterns;
}

/*
 * Adds an include/exclude pattern to the internal array.
 */
static void
add_epattern(char *pat, size_t len)
{

	/* Increase size if necessary */
	if (epatterns == epattern_sz) {
		epattern_sz *= 2;
		epattern = grep_realloc(epattern, ++epattern_sz * sizeof(*epattern));
	}
	if (len > 0 && pat[len - 1] == '\n')
		 --len;
	epattern[epatterns] = grep_malloc(len + 1);
	memcpy(epattern[epatterns], pat, len);
	epattern[epatterns][len] = '\0';
	++epatterns;
}

/*
 * Reads searching patterns from a file and adds them with add_pattern().
 */
static void
read_patterns(const char *fn)
{
	FILE	*f;
	char	*line;
	size_t	 len;

	if ((f = fopen(fn, "r")) == NULL)
		err(2, "%s", fn);
	while ((line = fgetln(f, &len)) != NULL)
		add_pattern(line, *line == '\n' ? 0 : len);
	if (ferror(f))
		err(2, "%s", fn);
	fclose(f);
}

int
main(int argc, char *argv[])
{
	int		 c, lastc, prevoptind, newarg, i, needpattern;
	char		*ep;
	unsigned long long l;

	setlocale(LC_ALL, "");

#ifndef WITHOUT_NLS
	catalog = catopen("grep", NL_CAT_LOCALE);
#endif

	/* Check what is the program name of the binary.  In this
	   way we can have all the funcionalities in one binary
	   without the need of scripting and using ugly hacks. */
	switch (__progname[0]) {
	case 'e':
		grepbehave = GREP_EXTENDED;
		break;
	case 'f':
		grepbehave = GREP_FIXED;
		break;
	case 'g':
		grepbehave = GREP_BASIC;
		break;
	case 'z':
		filebehave = FILE_GZIP;
		switch(__progname[1]) {
		case 'e':
			grepbehave = GREP_EXTENDED;
			break;
		case 'f':
			grepbehave = GREP_FIXED;
			break;
		case 'g':
			grepbehave = GREP_BASIC;
			break;
		}
		break;
	}

	lastc = '\0';
	newarg = 1;
	prevoptind = 1;
	needpattern = 1;
	while ((c = getopt_long(argc, argv, optstr,
				long_options, NULL)) != -1) {
		switch (c) {
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			if (newarg || !isdigit(lastc))
				Aflag = 0;
			else if (Aflag > LLONG_MAX / 10)
				errx(2, getstr(9));
			Aflag = Bflag = (Aflag * 10) + (c - '0');
			break;
		case 'C':
			if (optarg == NULL) {
				Aflag = Bflag = 2;
				break;
			}
			/* FALLTHROUGH */
		case 'A':
		case 'B':
			l = strtoull(optarg, &ep, 10);
			if ((errno == ERANGE) && (l == ULLONG_MAX))
				errx(2, getstr(9));
			if (c == 'A')
				Aflag = l;
			else if (c == 'B')
				Bflag = l;
			else
				Aflag = Bflag = l;
			break;
		case 'a':
			binbehave = BINFILE_TEXT;
			break;
		case 'b':
			bflag = 1;
			break;
		case 'c':
			cflag = 1;
			break;
		case 'D':
			if (strcmp(optarg, "skip") == 0)
				devbehave = DEV_SKIP;
			break;
		case 'd':
			if (strcmp("recurse", optarg) == 0) {
				Hflag++;
				dirbehave = DIR_RECURSE;
			} else if (strcmp("skip", optarg) == 0)
				dirbehave = DIR_SKIP;
			else if (strcmp("read", optarg) != 0)
				errx(2, getstr(13));
			break;
		case 'E':
			grepbehave = GREP_EXTENDED;
			break;
		case 'e':
			add_pattern(optarg, strlen(optarg));
			needpattern = 0;
			break;
		case 'F':
			grepbehave = GREP_FIXED;
			break;
		case 'f':
			read_patterns(optarg);
			needpattern = 0;
			break;
		case 'G':
			grepbehave = GREP_BASIC;
			break;
		case 'H':
			Hflag++;
			break;
		case 'h':
			Hflag = 0;
			hflag = 1;
			break;
		case 'I':
			binbehave = BINFILE_SKIP;
			break;
		case 'i':
		case 'y':
			iflag = 1;
			cflags |= REG_ICASE;
			break;
		case 'J':
			filebehave = FILE_BZIP;
			break;
		case 'L':
			lflag = 0;
			Lflag = qflag = 1;
			break;
		case 'l':
			Lflag = 0;
			lflag = qflag = 1;
			break;
		case 'm':
			mflag++;
			mcount = strtoull(optarg, (char **)NULL, 10);
			if ((errno == ERANGE) && (mcount == ULLONG_MAX))
				err(2, getstr(13));
			break;
		case 'n':
			nflag = 1;
			break;
		case 'O':
			linkbehave = LINK_EXPLICIT;
			break;
		case 'o':
			oflag++;
			break;
		case 'P':
			grepbehave = GREP_PERL;
			break;
		case 'p':
			linkbehave = LINK_SKIP;
			break;
		case 'q':
			qflag = 1;
			break;
		case 'S':
			linkbehave = LINK_GREP;
			break;
		case 'R':
		case 'r':
			dirbehave = DIR_RECURSE;
			Hflag++;
			break;
		case 's':
			sflag = 1;
			break;
		case 'U':
			binbehave = BINFILE_BIN;
			break;
		case 'u':
		case MMAP_OPT:
			/* noop, compatibility */
			break;
		case 'V':
			printf(getstr(10), __progname, VERSION);
			exit(0);
		case 'v':
			vflag = 1;
			break;
		case 'w':
			wflag = 1;
			break;
		case 'x':
			xflag = 1;
			break;
		case 'Z':
			filebehave = FILE_GZIP;
			break;
		case BIN_OPT:
			if (strcmp("binary", optarg) == 0)
				binbehave = BINFILE_BIN;
			else if (strcmp("without-match", optarg) == 0)
				binbehave = BINFILE_SKIP;
			else if (strcmp("text", optarg) == 0)
				binbehave = BINFILE_TEXT;
			else
				errx(2, getstr(11));
			break;
		case COLOR_OPT:
			if (optarg == NULL)
				optarg = "auto";
			if (strcmp("auto", optarg) == 0 || strcmp("always", optarg) == 0 ) {
				color = getenv("GREP_COLOR");
				if (color == NULL)
					color = "01;31";
			} else if (strcmp("never", optarg) == 0)
				color = NULL;
			else
				errx(2, getstr(15));
			break;
		case LABEL_OPT:
			label = optarg;
			break;
		case LINEBUF_OPT:
			lbflag = 1;
			break;
		case NULL_OPT:
			nullflag = 1;
			break;
		case R_INCLUDE_OPT:
			if (dirbehave != DIR_RECURSE)
				usage();
			inclflag = 1;
			exclflag = 0;
			add_epattern(basename(optarg), strlen(basename(optarg)));
			break;
		case R_EXCLUDE_OPT:
			if (dirbehave != DIR_RECURSE)
				usage();
			inclflag = 0;
			exclflag = 1;
			add_epattern(basename(optarg), strlen(basename(optarg)));
			break;
		case HELP_OPT:
		default:
			usage();
		}
		lastc = c;
		newarg = optind != prevoptind;
		prevoptind = optind;
	}
	argc -= optind;
	argv += optind;

	/* Fail if we don't have any pattern */
	if (argc == 0 && needpattern)
		usage();

	/* Process patterns from command line */
	if (argc != 0 && needpattern) {
		add_pattern(*argv, strlen(*argv));
		--argc;
		++argv;
	}

	switch (grepbehave) {
	case GREP_FIXED:
		cflags |= REG_NOSPEC;
		break;
	case GREP_BASIC:
		cflags |= REG_BASIC;
		break;
	case GREP_EXTENDED:
		cflags |= REG_EXTENDED;
		break;
	case GREP_PERL:
#ifndef WITH_PCRE
		errx(2, "PCRE is not enabled in this version of grep."
			"To enable this feature, please install libpcre and"
			"recompile grep with WITH_PCRE set.");
#endif
		break;
	default:
		/* NOTREACHED */
		usage();
	}
	if (grepbehave != GREP_PERL) {
		/* Compile regexes with regcomp() */
		r_pattern = grep_calloc(patterns, sizeof(*r_pattern));
		for (i = 0; i < patterns; ++i) {
			c = regcomp(&r_pattern[i], pattern[i], cflags);
			if (c != 0) {
				regerror(c, &r_pattern[i], re_error,
				    RE_ERROR_BUF);
				errx(2, "%s", re_error);
			}
		}
	} else {
#ifdef WITH_PCRE
		/* Compile Perl regexes with pcre_compile() */
		perl_pattern = grep_calloc(patterns, sizeof(perl_pattern));
		for (i = 0; i < patterns; ++i) {
			char	**err_msg = NULL;
			int	erroff;

			perl_pattern[i] = pcre_compile(pattern[i], 0, (const char **)err_msg, &erroff, NULL);
			if (perl_pattern[i] != NULL)
				errx(2, "wrong PCRE: %s", err_msg[0]);
		}
#else
		;
#endif
	}

	if (lbflag)
		setlinebuf(stdout);

	if ((argc == 0 || argc == 1) && !Hflag)
		hflag = 1;

	if (argc == 0)
		exit(!procfile("-"));

	if (dirbehave == DIR_RECURSE)
		c = grep_tree(argv);
	else 
		for (c = 0; argc--; ++argv)
			c+= procfile(*argv);

#ifndef WITHOUT_NLS
	catclose(catalog);
#endif

	/* Find out the correct return value according to the
	   results and the command line option. */
	if (c) {
		if (notfound && qflag)
			exit(0);
		else if (notfound)
			exit (2);
		else
			exit (0);		
	} else if (notfound)
		exit(2);
	else
		exit(1);
}
